import pooch
import pytest

from wpodnet import WPODNet, load_wpodnet_from_checkpoint


@pytest.fixture(scope="session")
def wpodnet() -> WPODNet:
    checkpoint = pooch.retrieve(
        "https://github.com/Pandede/WPODNet-Pytorch/releases/download/1.0.0/wpodnet.pth",
        known_hash="ac9fded54614d01b3082dd4e3917a65d4720b77a3f468fa934dbd85c814d3d77",
    )
    return load_wpodnet_from_checkpoint(checkpoint)



import torch


class BasicConvBlock(torch.nn.Module):
    def __init__(self, in_channels: int, out_channels: int):
        super().__init__()
        self.conv_layer = torch.nn.Conv2d(
            in_channels, out_channels, kernel_size=3, padding=1
        )
        self.bn_layer = torch.nn.BatchNorm2d(out_channels, momentum=0.99, eps=0.001)
        self.act_layer = torch.nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv_layer(x)
        x = self.bn_layer(x)
        return self.act_layer(x)


class ResBlock(torch.nn.Module):
    def __init__(self, channels: int):
        super().__init__()
        self.conv_block = BasicConvBlock(channels, channels)
        self.sec_layer = torch.nn.Conv2d(channels, channels, kernel_size=3, padding=1)
        self.bn_layer = torch.nn.BatchNorm2d(channels, momentum=0.99, eps=0.001)
        self.act_layer = torch.nn.ReLU(inplace=True)

    def forward(self, x):
        h = self.conv_block(x)
        h = self.sec_layer(h)
        h = self.bn_layer(h)
        return self.act_layer(x + h)


class WPODNet(torch.nn.Module):
    """
    WPODNet in PyTorch.

    The original architecture is built in Keras: https://github.com/sergiomsilva/alpr-unconstrained/blob/master/create-model.py
    """

    # https://github.com/sergiomsilva/alpr-unconstrained/blob/master/src/keras_utils.py#L43-L44
    stride = 16  # net_stride
    scale_factor = 7.75  # side

    def __init__(self):
        super().__init__()
        self.backbone = torch.nn.Sequential(
            BasicConvBlock(3, 16),
            BasicConvBlock(16, 16),
            torch.nn.MaxPool2d(2),
            BasicConvBlock(16, 32),
            ResBlock(32),
            torch.nn.MaxPool2d(2),
            BasicConvBlock(32, 64),
            ResBlock(64),
            ResBlock(64),
            torch.nn.MaxPool2d(2),
            BasicConvBlock(64, 64),
            ResBlock(64),
            ResBlock(64),
            torch.nn.MaxPool2d(2),
            BasicConvBlock(64, 128),
            ResBlock(128),
            ResBlock(128),
            ResBlock(128),
            ResBlock(128),
        )
        self.prob_layer = torch.nn.Conv2d(128, 2, kernel_size=3, padding=1)
        self.bbox_layer = torch.nn.Conv2d(128, 6, kernel_size=3, padding=1)

        # Registry a dummy tensor for retrieve the attached device
        self.register_buffer("dummy", torch.Tensor(), persistent=False)

    @property
    def device(self) -> torch.device:
        return self.dummy.device

    def forward(self, image: torch.Tensor):
        feature: torch.Tensor = self.backbone(image)
        probs: torch.Tensor = self.prob_layer(feature)
        probs = torch.softmax(probs, dim=1)
        affines: torch.Tensor = self.bbox_layer(feature)

        return probs, affines



from pathlib import Path
from typing import Union

import torch

from .backend import Prediction, Predictor
from .model import WPODNet


def load_wpodnet_from_checkpoint(ckpt_path: Union[str, Path]) -> WPODNet:
    """
    Load a pre-trained WPOD-NET model from a checkpoint file.

    Args:
        ckpt_path (Union[str, Path]): The path to the checkpoint file.

    Returns:
        WPODNet: The WPOD-NET model with pretrained weights loaded from the checkpoint.
    """
    model = WPODNet()

    # Load the state dictionary from the checkpoint
    checkpoint = torch.load(ckpt_path, weights_only=True)
    model.load_state_dict(checkpoint)

    return model


__all__ = ["Prediction", "Predictor", "WPODNet", "load_wpodnet_from_checkpoint"]



import errno
from argparse import ArgumentParser, ArgumentTypeError
from pathlib import Path
from typing import List, Union

import torch
from PIL import Image, UnidentifiedImageError

from wpodnet import Predictor, load_wpodnet_from_checkpoint


def list_image_paths(p: Union[str, Path]) -> List[Path]:
    """
    List all images in a directory.

    Args:
        path (Union[str, Path]): The path to the directory containing images.

    Returns:
        Generator[Image.Image]: A generator of PIL Image objects.
    """
    p = Path(p)
    if not p.is_dir():
        raise FileNotFoundError(errno.ENOTDIR, "No such directory", args.save_annotated)

    image_paths: List[Path] = []
    for f in p.glob("**/*"):
        try:
            with Image.open(f) as image:
                image.verify()
            image_paths.append(f)
        except UnidentifiedImageError:
            pass
    return image_paths


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("source", type=str, help="the path to the image")
    parser.add_argument(
        "-w", "--weight", type=str, required=True, help="the path to the model weight"
    )
    parser.add_argument(
        "--scale",
        type=float,
        default=1.0,
        help="adjust the scaling ratio. default to 1.0.",
    )
    parser.add_argument(
        "--save-annotated",
        type=str,
        help="save the annotated image at the given folder",
    )
    parser.add_argument(
        "--save-warped", type=str, help="save the warped image at the given folder"
    )
    args = parser.parse_args()

    if args.scale <= 0.0:
        raise ArgumentTypeError(message="scale must be greater than 0.0")

    if args.save_annotated is not None:
        save_annotated = Path(args.save_annotated)
        if not save_annotated.is_dir():
            raise FileNotFoundError(
                errno.ENOTDIR, "No such directory", args.save_annotated
            )
    else:
        save_annotated = None

    if args.save_warped is not None:
        save_warped = Path(args.save_warped)
        if not save_warped.is_dir():
            raise FileNotFoundError(
                errno.ENOTDIR, "No such directory", args.save_warped
            )
    else:
        save_warped = None

    # Prepare for the model
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model = load_wpodnet_from_checkpoint(args.weight).to(device)

    predictor = Predictor(model)

    source = Path(args.source)
    if source.is_file():
        image_paths = [source]
    elif source.is_dir():
        image_paths = list_image_paths(source)
    else:
        raise FileNotFoundError(errno.ENOENT, "No such file or directory", args.source)

    for i, image_path in enumerate(image_paths):
        image = Image.open(image_path)
        prediction = predictor.predict(image, scaling_ratio=args.scale)

        print(f"Prediction #{i}")
        print("  bounds", prediction.bounds)
        print("  confidence", prediction.confidence)

        if save_annotated:
            annotated_path = save_annotated / Path(image.filename).name

            canvas = image.copy()
            prediction.annotate(canvas, outline="red")
            canvas.save(annotated_path)
            print(f"Saved the annotated image at {annotated_path}")

        if save_warped:
            warped_path = save_warped / Path(image.filename).name
            warped = prediction.warp(image)
            warped.save(warped_path)
            print(f"Saved the warped image at {warped_path}")

        print()



from typing import List, Tuple

import pytest
from PIL import Image

from wpodnet import Predictor, WPODNet


@pytest.fixture
def predictor(wpodnet: WPODNet) -> Predictor:
    return Predictor(wpodnet)


class TestPredictor:
    @pytest.mark.parametrize(
        "image_path, bounds",
        [
            (
                "docs/sample/original/03009.jpg",
                [(1384, 682), (1517, 614), (1502, 685), (1369, 753)],
            ),
            (
                "docs/sample/original/03016.jpg",
                [(567, 375), (643, 368), (643, 393), (567, 400)],
            ),
            (
                "docs/sample/original/03025.jpg",
                [(94, 162), (162, 171), (160, 186), (92, 177)],
            ),
        ],
    )
    def test_predict(
        self, predictor: Predictor, image_path: str, bounds: List[Tuple[int, int]]
    ):
        image = Image.open(image_path)
        prediction = predictor.predict(image)
        assert prediction.bounds == bounds
        assert prediction.confidence >= 0.9



from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np
import torch
from PIL import Image, ImageDraw
from torchvision.transforms.functional import _get_perspective_coeffs, to_tensor

from .model import WPODNet


@dataclass(frozen=True)
class Prediction:
    """
    The prediction result from WPODNet.

    Attributes:
        bounds (List[Tuple[int, int]]): The bounding coordinates of the detected license plate. Must be a list of 4 points (x, y).
        confidence (float): The confidence score of the detection. Must be between 0.0 and 1.0.
    """

    bounds: List[Tuple[int, int]]
    confidence: float

    def __post_init__(self):
        if len(self.bounds) != 4:
            raise ValueError(
                f"expected bounds to have 4 points, got {len(self.bounds)} points"
            )
        if self.confidence < 0 or self.confidence > 1:
            raise ValueError(
                f"confidence must be between 0.0 and 1.0, got {self.confidence}"
            )

    def annotate(
        self,
        canvas: Image.Image,
        fill: Optional[str] = None,
        outline: Optional[str] = None,
        width: int = 1,
    ) -> None:  # pragma: no cover
        """
        Annotates the image with the bounding polygon.

        Args:
            canvas (PIL.Image.Image): The image to be annotated.
            fill (Optional[str]): The fill color for the polygon. Defaults to None.
            outline (Optional[str]): The outline color for the polygon. Defaults to None.
            width (int): The width of the outline. Defaults to 1.

        Note:
            The arguments `fill`, `outline`, and `width` are passed to the `ImageDraw.Draw.polygon` method.
            See https://pillow.readthedocs.io/en/stable/reference/ImageDraw.html#PIL.ImageDraw.ImageDraw.polygon.
        """
        drawer = ImageDraw.Draw(canvas)
        drawer.polygon(self.bounds, fill=fill, outline=outline, width=width)

    def warp(self, canvas: Image.Image) -> Image.Image:  # pragma: no cover
        """
        Warps the image with perspective based on the bounding polygon.

        Args:
            canvas (PIL.Image.Image): The image to be warped.

        Returns:
            PIL.Image.Image: The warped image.
        """
        coeffs = _get_perspective_coeffs(
            startpoints=self.bounds,
            endpoints=[
                (0, 0),
                (canvas.width, 0),
                (canvas.width, canvas.height),
                (0, canvas.height),
            ],
        )
        return canvas.transform(
            (canvas.width, canvas.height), Image.Transform.PERSPECTIVE, coeffs
        )


Q = np.array(
    [
        [-0.5, 0.5, 0.5, -0.5],
        [-0.5, -0.5, 0.5, 0.5],
        [1.0, 1.0, 1.0, 1.0],
    ]
)


class Predictor:
    """A wrapper class for WPODNet to make predictions."""

    def __init__(self, wpodnet: WPODNet) -> None:
        """
        Args:
            wpodnet (WPODNet): The WPODNet model to use for prediction.
        """
        self.wpodnet = wpodnet
        self.wpodnet.eval()

    def _resize_to_fixed_ratio(
        self, image: Image.Image, dim_min: int, dim_max: int
    ) -> Image.Image:
        h, w = image.height, image.width

        wh_ratio = max(h, w) / min(h, w)
        side = int(wh_ratio * dim_min)
        bound_dim = min(side + side % self.wpodnet.stride, dim_max)

        factor = bound_dim / max(h, w)
        reg_w, reg_h = int(w * factor), int(h * factor)

        # Ensure the both width and height are the multiply of `self.wpodnet.stride`
        reg_w_mod = reg_w % self.wpodnet.stride
        if reg_w_mod > 0:
            reg_w += self.wpodnet.stride - reg_w_mod

        reg_h_mod = reg_h % self.wpodnet.stride
        if reg_h_mod > 0:
            reg_h += self.wpodnet.stride - reg_h_mod

        return image.resize((reg_w, reg_h))

    def _to_torch_image(self, image: Image.Image) -> torch.Tensor:
        tensor = to_tensor(image)
        return tensor.unsqueeze_(0)

    def _inference(self, image: torch.Tensor) -> Tuple[np.ndarray, np.ndarray]:
        with torch.no_grad():
            probs, affines = self.wpodnet.forward(image)

        # Convert to squeezed numpy array
        # grid_w: The number of anchors in row
        # grid_h: The number of anchors in column
        probs = np.squeeze(probs.cpu().numpy())[0]  # (grid_h, grid_w)
        affines = np.squeeze(affines.cpu().numpy())  # (6, grid_h, grid_w)

        return probs, affines

    def _get_max_anchor(self, probs: np.ndarray) -> Tuple[int, int]:
        return np.unravel_index(probs.argmax(), probs.shape)

    def _get_bounds(
        self,
        affines: np.ndarray,
        anchor_y: int,
        anchor_x: int,
        scaling_ratio: float = 1.0,
    ) -> np.ndarray:
        # Compute theta
        theta = affines[:, anchor_y, anchor_x]
        theta = theta.reshape((2, 3))
        theta[0, 0] = max(theta[0, 0], 0.0)
        theta[1, 1] = max(theta[1, 1], 0.0)

        # Convert theta into the bounding polygon
        bounds = np.matmul(theta, Q) * self.wpodnet.scale_factor * scaling_ratio

        # Normalize the bounds
        _, grid_h, grid_w = affines.shape
        bounds[0] = (bounds[0] + anchor_x + 0.5) / grid_w
        bounds[1] = (bounds[1] + anchor_y + 0.5) / grid_h

        return np.transpose(bounds)

    def predict(
        self,
        image: Image.Image,
        scaling_ratio: float = 1.0,
        dim_min: int = 512,
        dim_max: int = 768,
    ) -> Prediction:
        """
        Detect license plate in the image.

        Args:
            image (Image.Image): The image to be detected.
            scaling_ratio (float): The scaling ratio of the resulting bounding polygon. Default to 1.0.
            dim_min (int): The minimum dimension of the resized image. Default to 512
            dim_max (int): The maximum dimension of the resized image. Default to 768

        Returns:
            Prediction: The prediction result with highest confidence.
        """
        orig_h, orig_w = image.height, image.width

        # Resize the image to fixed ratio
        # This operation is convienence for setup the anchors
        resized = self._resize_to_fixed_ratio(image, dim_min=dim_min, dim_max=dim_max)
        resized = self._to_torch_image(resized)
        resized = resized.to(self.wpodnet.device)

        # Inference with WPODNet
        # probs: The probability distribution of the location of license plate
        # affines: The predicted affine matrix
        probs, affines = self._inference(resized)

        # Get the theta with maximum probability
        max_prob = np.amax(probs)
        anchor_y, anchor_x = self._get_max_anchor(probs)
        bounds = self._get_bounds(affines, anchor_y, anchor_x, scaling_ratio)

        bounds[:, 0] *= orig_w
        bounds[:, 1] *= orig_h

        return Prediction(
            bounds=[(x, y) for x, y in np.int32(bounds).tolist()],
            confidence=max_prob.item(),
        )



import pytest

from wpodnet import Prediction


class TestPrediction:
    def test_validation(self):
        # Valid
        Prediction(bounds=[(1, 2), (3, 4), (5, 6), (7, 8)], confidence=0.95)

        # Invalid bounds length
        with pytest.raises(ValueError):
            Prediction(bounds=[(1, 2), (3, 4), (5, 6)], confidence=0.95)

        # Invalid confidence value
        with pytest.raises(ValueError):
            Prediction(bounds=[(1, 2), (3, 4), (5, 6), (7, 8)], confidence=1.01)
